/*
** ADM-XRC Series Driver
**
** (C) Copyright 2001-2003 Alpha Data
**
** flash.c - Flash memory access routines
**
** ----------------------------------------------------------------------------
*/

#include "device.h"
#include "flash.h"

static MethodResult
writeBlock(
  DeviceContext*  pDevCtx,
  os_uint32       start,
  os_uint32       length,
  const os_uint8* data)
{
  MethodResult	result = MethodResult_Success;
  os_uint16     venDev;
  os_uint32     i, limit;
  unsigned long timeout_us, remaining_us;
  os_boolean    bTimeout;
  os_uint8      val8, byte;
  
  os_dbgprint(3, ("writeBlock: entered, start=0x%lx length=0x%lx data=0x%p\n",
		  (unsigned long) start, (unsigned long) length, data));

  if (!pDevCtx->info.flash.present) {
    return MethodResult_NotPresent;
  }

  limit = start + length;

  venDev = pDevCtx->info.flash.identity.vendor;
  venDev = (venDev << 8) | pDevCtx->info.flash.identity.device;
  switch (venDev) {
  case 0x8999: /* Micron 28F008B */
    {
      for (i = start; i < limit; i++) {
	byte = *data++;
	/* Send WRITE-BYTE command */
	val8 = 0x40U;
	pDevCtx->methods.flash(pDevCtx, TRUE, 0x0, &val8);
	val8 = byte;
	pDevCtx->methods.flash(pDevCtx, TRUE, i, &val8);

	/* Wait for completion */
	timeout_us = 10;
	bTimeout = FALSE;
	while (timeout_us) {
	  os_delay_busy_us(2);
	  /* Use data polling method */
	  pDevCtx->methods.flash(pDevCtx, FALSE, i, &val8);
	  if (val8 & 0x80U) {
	    break;
	  }
	  if (timeout_us < 2) {
	    bTimeout = TRUE;
	    break;
	  }
	  timeout_us -= 2;
	}
	if (bTimeout) {
	  result = MethodResult_Timeout;
	} else {
	  /* Status-check */
	  if (val8 & 0x0cU) {
	    os_dbgprint(0, ("*** writeBlock: Error detected after WRITE-BYTE: SR[7:0] = 0x%lx\n",
			    (unsigned long) val8));
	    result = MethodResult_VerifyFailed;
	  }
	}
	
	/* Return to READ-ARRAY mode */
	val8 = 0xffU;
	pDevCtx->methods.flash(pDevCtx, TRUE, 0x0, &val8);
      }
      break;
    }
    
  case 0x8916: /* Intel 28320J3A */
  case 0x8917: /* Intel 28640J3A */
  case 0x8918: /* Intel 28128J3A */
    {
      timeout_us = 2000;
      i = start;
      while (length) {
	os_uint32 j, n;
	
	/* Calculate number of bytes for this command */
	n = 32 - (i & 0x1fU);
	if (n > length) {
	  n = length;
	}
	
	os_dbgprint(3, ("writeBlock: doing WRITE-TO-BUFFER, start=0x%lx length=0x%lx i=0x%lx n=0x%lx data=0x%p\n",
			(unsigned long) start, (unsigned long) length, (unsigned long) i, (unsigned long) n, data));

	/* Send WRITE-TO-BUFFER command */
	val8 = 0xe8U;
	pDevCtx->methods.flash(pDevCtx, TRUE, start, &val8);
	
	/* Send the data to the write buffer */
	val8 = (os_uint8) (n - 1);
	pDevCtx->methods.flash(pDevCtx, TRUE, start, &val8);
	for (j = 0; j < n; j++) {
	  val8 = *data++;
	  pDevCtx->methods.flash(pDevCtx, TRUE, i + j, &val8);
	}
	
	/* Confirm WRITE-TO-BUFFER command */
	val8 = 0xd0U;
	pDevCtx->methods.flash(pDevCtx, TRUE, start, &val8);
	
	/* Wait for completion of WRITE-TO-BUFFER command */
	remaining_us = timeout_us;
	bTimeout = FALSE;
	while (1) {
	  /* Send Read-STATUS command */
	  val8 = 0x70U;
	  pDevCtx->methods.flash(pDevCtx, TRUE, start, &val8);
	  pDevCtx->methods.flash(pDevCtx, FALSE, start, &val8);
	  if (val8 & 0x80U) {
	    break;
	  }

	  os_delay_busy_us(10);
	  if (remaining_us < 10) {
	    os_dbgprint(0, ("*** writeBlock: timeout waiting for buffered write to complete\n"));
	    bTimeout = TRUE;
	    break;
	  }
	  remaining_us -= 10;
	}
	
	if (bTimeout) {
	  /* Send CLEAR-STATUS command */
	  val8 = 0x50U;
	  pDevCtx->methods.flash(pDevCtx, TRUE, start, &val8);
	  /* Send READ-ARRAY command */
	  val8 = 0xffU;
	  pDevCtx->methods.flash(pDevCtx, TRUE, start, &val8);
	  result = MethodResult_Timeout;
	  /* Return early */
	  break;
	}
	
	/* Perform status-check */
	if (val8 & 0x0cU) {
	  os_dbgprint(0, ("*** writeBlock: Error detected after WRITE-BYTE: SR[7:0] = 0x%lx\n",
			  (unsigned long) val8));
	  result = MethodResult_VerifyFailed;
	  /* Send CLEAR-STATUS command */
	  val8 = 0x50U;
	  pDevCtx->methods.flash(pDevCtx, TRUE, start, &val8);
	  /* Send READ-ARRAY command */
	  val8 = 0xffU;
	  pDevCtx->methods.flash(pDevCtx, TRUE, start, &val8);
	  /* Return early */
	  break;
	}
	
	/* Send READ-ARRAY command */
	val8 = 0xffU;
	pDevCtx->methods.flash(pDevCtx, TRUE, 0x0, &val8);

	i += n;
	length -= n;
      }
      break;
    }
    
  case 0xbfd9: /* SST 39VF016Q */
    {
      os_uint8 byte;
      
      for (i = start; i < limit; i++) {
	byte = *data++;
	/* Send WRITE-BYTE command */
	val8 = 0xaaU;
	pDevCtx->methods.flash(pDevCtx, TRUE, 0x5555, &val8);
	val8 = 0x55U;
	pDevCtx->methods.flash(pDevCtx, TRUE, 0x2aaa, &val8);
	val8 = 0xa0U;
	pDevCtx->methods.flash(pDevCtx, TRUE, 0x5555, &val8);
	val8 = byte;
	pDevCtx->methods.flash(pDevCtx, TRUE, i, &val8);
	
	/* Wait for completion */
	timeout_us = 25;
	bTimeout = FALSE;
	while (timeout_us) {
	  os_delay_busy_us(5);
	  /* Use data polling method */
	  pDevCtx->methods.flash(pDevCtx, FALSE, i, &val8);
	  if ((val8 & 0x80U) == (byte & 0x80U)) {
	    break;
	  }
	  if (timeout_us < 5) {
	    bTimeout = TRUE;
	    break;
	  }
	  timeout_us -= 5;
	}
	if (bTimeout) {
	  result = MethodResult_Timeout;
	  break;
	}
	
	/* Return to READ-ARRAY mode */
	val8 = 0xf0U;
	pDevCtx->methods.flash(pDevCtx, TRUE, 0x0, &val8);
      }
      break;
    }
    
  default:
    os_assert(FALSE);
    result = MethodResult_NotSupported;
    break;
  }
  
  return result;
}

/*
** -----------------------------------------------------------------
** Exported routines
** -----------------------------------------------------------------
*/

/*
** Capabilities of supported Flash devices
**
** Device        Ven:Dev  Size    Block-erase   Bootblocks
**
** 28F0008B       89:99   4M      128k, 14s     16k, 8k, 8k, 96k
** 28F320         89:16   4M      128k, 5s      -
** 28F640         89:17   8M      128k, 5s      -
** 28F128         89:18   16M     128k, 5s      -
** 39VF016Q       bf:d9   2M      64k, 25ms     -
*/
void
identifyFlash(
  DeviceContext* pDevCtx)
{
  os_boolean   bFound = FALSE;
  os_boolean   res;
  os_uint8     val8;
  os_uint8     vendor = 0xffU, device = 0xffU;
  os_uint32    size = 0;
  os_uint32    eraseBlock = 0;
  unsigned int numBootBlock = 0;
  os_uint32    bootBlock[4];
  char*        name = "unknown";

  /* Go to read array mode */
  val8 = 0xffU;
  res = pDevCtx->methods.flash(pDevCtx, TRUE, 0, &val8);
  if (res != MethodResult_Success) {
    return;
  }
  val8 = 0xf0U;
  res = pDevCtx->methods.flash(pDevCtx, TRUE, 0, &val8);
  
  if (!bFound) {
    /*
    ** Try Intel.
    **
    ** Send "Read Identifier Codes" command.
    */
    val8 = 0x90U;
    res = pDevCtx->methods.flash(pDevCtx, TRUE, 0, &val8);
    res = pDevCtx->methods.flash(pDevCtx, FALSE, 0, &val8);
    if (val8 == 0x89) {
      /* Identified Intel vendor - now get device */
      vendor = val8;
      res = pDevCtx->methods.flash(pDevCtx, FALSE, 0x2, &val8);
      device = val8;
      
      /* Go to read array mode */
      val8 = 0xffU;
      res = pDevCtx->methods.flash(pDevCtx, TRUE, 0, &val8);
      
      switch (device) {
      case 0x16:
	{
	  /* Intel 28F320J3A, 16Mbit */
	  bFound = TRUE;
	  eraseBlock = 128*1024;
	  size = 4*1024*1024;
	  name = "Intel 28F320";
	  break;
	}

      case 0x17:
	{
	  /* Intel 28F640J3A, 64Mbit */
	  bFound = TRUE;
	  eraseBlock = 128*1024;
	  size = 8*1024*1024;
	  name = "Intel 28F640";
	  break;
	}

      case 0x18:
	{
	  /* Intel 28F128J3A, 128Mbit */
	  bFound = TRUE;
	  eraseBlock = 128*1024;
	  size = 16*1024*1024;
	  name = "Intel 28F128";
	  break;
	}

      default:
	break;
      }
    }
  }

  if (!bFound) {
    /*
    ** Try Micron
    **
    ** Send "Read Identifier Codes" command
    */
    val8 = 0x90U;
    res = pDevCtx->methods.flash(pDevCtx, TRUE, 0, &val8);
    res = pDevCtx->methods.flash(pDevCtx, FALSE, 0, &val8);
    if (val8 == 0x89) {
      /* Vendor is Micron - now get device */
      vendor = 0x89;
      res = pDevCtx->methods.flash(pDevCtx, FALSE, 1, &val8);
      device = val8;
      
      /* Go to read array mode */
      val8 = 0xffU;
      res = pDevCtx->methods.flash(pDevCtx, TRUE, 0, &val8);
      
      switch (device) {
      case 0x99:
	{
	  /* Micron 28F008B, 8Mbit */
	  bFound = TRUE;
	  eraseBlock = 128*1024;
	  numBootBlock = 4;
	  bootBlock[0] = 16*1024;
	  bootBlock[1] = 8*1024;
	  bootBlock[2] = 8*1024;
	  bootBlock[3] = 96*1024;
	  size = 1*1024*1024;
	  name = "Micron 28F008B";
	  break;
	}

      default:
	break;
      }
    }
  }
  
  if (!bFound) {
    /*
    ** Try SST/ATMEL - these devices conform to an older JEDEC standard.
    **
    ** Enter "Read ID" mode.
    */
    val8 = 0xaaU;
    res = pDevCtx->methods.flash(pDevCtx, TRUE, 0x5555, &val8);
    val8 = 0x55U;
    res = pDevCtx->methods.flash(pDevCtx, TRUE, 0xaaaa, &val8);
    val8 = 0x90U;
    res = pDevCtx->methods.flash(pDevCtx, TRUE, 0x5555, &val8);
    res = pDevCtx->methods.flash(pDevCtx, FALSE, 0, &val8);
    if (val8 == 0xbfU) {
      vendor = val8;
      res = pDevCtx->methods.flash(pDevCtx, FALSE, 1, &val8);
      device = val8;

      /* Exit read identity mode */
      val8 = 0xf0U;
      res = pDevCtx->methods.flash(pDevCtx, TRUE, 0, &val8);
      
      switch (device) {
      case 0xd9U:
	{
	  /* SST 39VF016Q, 16Mbit */
	  bFound = TRUE;
	  eraseBlock = 64*1024;
	  size = 2*1024*1024;
	  name = "SST 39VF016Q";
	  break;
	}

      default:
	break;
      }
    }
    
    /* Go to read array mode */
    val8 = 0xffU;
    res = pDevCtx->methods.flash(pDevCtx, TRUE, 0, &val8);
    val8 = 0xf0U;
    res = pDevCtx->methods.flash(pDevCtx, TRUE, 0, &val8);
  }
  
  if (bFound) {
    unsigned int	i;
    
    os_dbgprint(1, ("identifyFlash: identified Flash as %s, eraseBlock = 0x%lx\n",
		    name, (unsigned long) eraseBlock));
    pDevCtx->info.flash.present = TRUE;
    pDevCtx->info.flash.identity.device = device;
    pDevCtx->info.flash.identity.vendor = vendor;
    pDevCtx->info.flash.identity.name = name;
    pDevCtx->info.flash.size = size;
    pDevCtx->info.flash.layout.numBootBlock = numBootBlock;
    for (i = 0; i < numBootBlock; i++) {
      pDevCtx->info.flash.layout.bootBlock[i] = bootBlock[i];
    }
    pDevCtx->info.flash.layout.eraseBlock = eraseBlock;
    
    pDevCtx->info.flash.targetArea.start = 0;
    pDevCtx->info.flash.targetArea.length = 0;
    switch (pDevCtx->devDesc.id.pci.subDId) {
    case PCI_SUBDEVICE_ADATA_ADMXRC:
    case PCI_SUBDEVICE_ADATA_ADMXRC2L:
    case PCI_SUBDEVICE_ADATA_ADMXRC2:
    case PCI_SUBDEVICE_ADATA_ADPWRC2:
    case PCI_SUBDEVICE_ADATA_ADPDRC2:
      pDevCtx->info.flash.targetArea.start = 0;
      pDevCtx->info.flash.targetArea.length = size;
      break;
      
    case PCI_SUBDEVICE_ADATA_ADMXRC2PL_PCIX:
    case PCI_SUBDEVICE_ADATA_ADMXRC2PL_PCI:
      pDevCtx->info.flash.targetArea.start = 0x800000U;
      pDevCtx->info.flash.targetArea.length = size - 0x800000U;
      break;

    case PCI_SUBDEVICE_ADATA_ADMXRC2P:
      pDevCtx->info.flash.targetArea.start = 0x40000U;
      pDevCtx->info.flash.targetArea.length = size - 0x40000U;
      break;

    default:
      os_assert(FALSE);
      break;
    }
  } else {
    os_dbgprint(0, ("*** identifyFlash: failed to identify Flash\n"));
  }
}

MethodResult
identifyFlashBlock(
  DeviceContext* pDevCtx,
  os_uint32      location,
  os_uint32*     blockStart,
  os_uint32*     blockSize,
  os_boolean*    isBootBlock)
{
  os_uint32    eraseBlock = pDevCtx->info.flash.layout.eraseBlock;
  os_uint32    start, end, size = 0;
  os_boolean   bBoot;
  unsigned int i;
  
  if (location > pDevCtx->info.flash.size) {
    return MethodResult_InvalidLocation;
  }
  
  /* Identify which block the location belongs to */
  start = 0;
  bBoot = FALSE;
  for (i = 0; i < pDevCtx->info.flash.layout.numBootBlock; i++) {
    size = pDevCtx->info.flash.layout.bootBlock[i];
    end = start + size;
    if (location >= start && location < end) {
      bBoot = TRUE;
      break;
    }
    start += size;
  }
  
  if (!bBoot) {
    /* It's not a boot block, must be a normal block */
    start = location & ~(eraseBlock - 1);
    size = eraseBlock;
  }
  
  *blockStart = start;
  *blockSize = size;
  *isBootBlock = bBoot;
  
  return MethodResult_Success;
}

MethodResult
eraseFlash(
  DeviceContext* pDevCtx,
  os_uint32      start,
  os_uint32      length)
{
  MethodResult  result = MethodResult_Success;
  os_uint32     eraseBlock;
  os_uint32     blockStart, blockEnd;
  os_uint32     blockSize;
  os_boolean    isBootBlock;
  os_uint32     chunk;
  os_uint16     venDev;
  os_uint8      val8;
  unsigned long	timeout_ms, remaining_ms;
  os_boolean    bTimeout, bInterrupt, bStatusError;
  unsigned long interval;

  os_dbgprint(3, ("eraseFlash: entered, start=0x%lx length=0x%lx\n",
		  (unsigned long) start, (unsigned long) length));

  if (!pDevCtx->info.flash.present) {
    return MethodResult_NotPresent;
  }
  
  venDev = pDevCtx->info.flash.identity.vendor;
  venDev = (venDev << 8) | pDevCtx->info.flash.identity.device;
  eraseBlock = pDevCtx->info.flash.layout.eraseBlock;
  
  switch (venDev) {
  case 0x8999: /* Micron 28F008B */
    timeout_ms = 14000;
    break;
    
  case 0x8916: /* Intel 28320J3A */
  case 0x8917: /* Intel 28640J3A */
  case 0x8918: /* Intel 28128J3A */
    timeout_ms = 5000;
    break;
    
  case 0xbfd9: /* SST 39VF016Q */
    timeout_ms = 25;
    break;
    
  default:
    /* Unrecognised device */
    return MethodResult_NotSupported;
  }
  
  while (length) {
    /* Identify which block the range belongs to */
    identifyFlashBlock(pDevCtx, start, &blockStart, &blockSize, &isBootBlock);
    blockEnd = blockStart + eraseBlock;
    chunk = blockEnd - start;
    if (chunk > length) {
      chunk = length;
    }
    
    os_dbgprint(6, ("eraseFlash: start=0x%lx length=0x%lx blockStart=0x%lx blockEnd=0x%lx blockSize=0x%lx chunk=0x%lx\n",
		    (unsigned long) start, (unsigned long) length, (unsigned long) blockStart,
		    (unsigned long) blockEnd, (unsigned long) blockSize, (unsigned long) chunk));
    
    /* Erase the block */
    switch (venDev) {
    case 0x8999: /* Micron 28F008B */
    case 0x8916: /* Intel 28320J3A */
    case 0x8917: /* Intel 28640J3A */
    case 0x8918: /* Intel 28128J3A */
      {
	/* Send ERASE-BLOCK command */
	val8 = 0x20U;
	pDevCtx->methods.flash(pDevCtx, TRUE, 0, &val8);
	val8 = 0xd0U;
	pDevCtx->methods.flash(pDevCtx, TRUE, blockStart, &val8);
	
	/* Wait for completion */
	remaining_ms = timeout_ms;
	bTimeout = FALSE;
	bStatusError = FALSE;
	bInterrupt = FALSE;
	while (1) {
	  pDevCtx->methods.flash(pDevCtx, FALSE, 0, &val8);
	  if (val8 & 0x80U) {
	    break;
	  }
	  interval = 5;
	  bInterrupt = os_delay_thread(FALSE, &interval);
	  if (bInterrupt) {
	    os_dbgprint(1, ("*** eraseFlash: delay interrupted\n"));
	    break;
	  }
	  if (remaining_ms < 5) {
	    bTimeout = TRUE;
	    break;
	  }
	  remaining_ms -= 5;
	}
	
	if (bInterrupt) {
	  result = MethodResult_GeneralFailure;
	} else {
	  if (bTimeout) {
	    result = MethodResult_Timeout;
	  } else {
	    /* Status-check */
	    if (val8 & 0x1cU) {
	      /* Vpp Error */
	      os_dbgprint(0, ("*** eraseFlash: Error detected after ERASE-BLOCK: SR[7:0] = 0x%lx\n",
			      (unsigned long) val8));
	      result = MethodResult_VerifyFailed;
	      bStatusError = TRUE;
	    }
	  }
	  
	  /* If an error occurred, attempt to return to normal operation */
	  if (bTimeout || bInterrupt || bStatusError) {
	    /* Send CLEAR-STATUS command */
	    val8 = 0x50U;
	    pDevCtx->methods.flash(pDevCtx, TRUE, start, &val8);
	  }
	}
	
	/* Send READ-ARRAY command */
	val8 = 0xffU;
	pDevCtx->methods.flash(pDevCtx, TRUE, blockStart, &val8);
	break;
      }
      
    case 0xbfd9: /* SST 39VF016Q */
      {
	os_uint32     i;
	unsigned long interval;
	
	/* Send ERASE-BLOCK command */
	val8 = 0xaaU;
	pDevCtx->methods.flash(pDevCtx, TRUE, 0x5555, &val8);
	val8 = 0x55U;
	pDevCtx->methods.flash(pDevCtx, TRUE, 0x2aaa, &val8);
	val8 = 0x80U;
	pDevCtx->methods.flash(pDevCtx, TRUE, 0x5555, &val8);
	val8 = 0xaaU;
	pDevCtx->methods.flash(pDevCtx, TRUE, 0x5555, &val8);
	val8 = 0x55U;
	pDevCtx->methods.flash(pDevCtx, TRUE, 0x2aaa, &val8);
	val8 = 0x50U;
	pDevCtx->methods.flash(pDevCtx, TRUE, blockStart, &val8);
	
	/* Wait for completion */
	interval = timeout_ms;
	bInterrupt = os_delay_thread(FALSE, &interval);
	
	/* Return to READ-ARRAY mode */
	val8 = 0xffU;
	pDevCtx->methods.flash(pDevCtx, TRUE, 0x0, &val8);
	
	/* Verify that sector has been erased */
	for (i = blockStart; i < blockEnd; i++) {
	  pDevCtx->methods.flash(pDevCtx, FALSE, i, &val8);
	  if (val8 != 0xff) {
	    result = MethodResult_VerifyFailed;
	    break;
	  }
	}
	
	break;
      }
    }
    
    if (result != MethodResult_Success) {
      return result;
    }
    
    length -= chunk;
    start += chunk;
  }
  
  return MethodResult_Success;
}

MethodResult
readFlash(
  DeviceContext* pDevCtx,
  os_uint32      start,
  os_uint32      length,
  void*          buffer)
{
  MethodResult result = MethodResult_Success;
  os_uint32    i, limit;
  os_uint8     val8;

  os_dbgprint(3, ("readFlash: entered, start=0x%lx length=0x%lx buffer=0x%p\n",
		  (unsigned long) start, (unsigned long) length, buffer));

  limit = start + length;
  for (i = start; i < limit; i++) {
    result = pDevCtx->methods.flash(pDevCtx, FALSE, i, &val8);
    if (result != MethodResult_Success) {
      break;
    }
    *((os_uint8*) buffer)++ = val8;
  }
  
  return result;
}

MethodResult
writeFlash(
  DeviceContext* pDevCtx,
  os_uint32      start,
  os_uint32      length,
  const void*    data)
{
  MethodResult result = MethodResult_Success;
  os_uint8*    saveBuff;
  os_uint32    eraseBlock;
  os_uint32    blockStart, blockEnd;
  os_uint32    blockSize;
  os_boolean   isBootBlock;

  os_dbgprint(3, ("writeFlash: entered, start=0x%lx length=0x%lx data=0x%p\n",
		  (unsigned long) start, (unsigned long) length, data));

  if (!pDevCtx->info.flash.present) {
    return MethodResult_NotPresent;
  }
  
  saveBuff = pDevCtx->info.flash.buffer;
  eraseBlock = pDevCtx->info.flash.layout.eraseBlock;

  while (length) {
    os_uint32	chunk;
    
    identifyFlashBlock(pDevCtx, start, &blockStart, &blockSize, &isBootBlock);
    blockEnd = blockStart + blockSize;
    chunk = blockEnd - start;
    if (chunk > length) {
      chunk = length;
    }
    
    os_dbgprint(6, ("writeFlash: start=0x%lx blockStart=0x%lx blockEnd=0x%lx blockSize=0x%lx chunk=0x%lx data=0x%p saveBuff=0x%p\n",
		    (unsigned long) start, (unsigned long) blockStart, (unsigned long) blockEnd,
		    (unsigned long) blockSize, (unsigned long) chunk, data, saveBuff));
    
    /* Read the block */
    result = readFlash(pDevCtx, blockStart, blockSize, saveBuff);
    if (result != MethodResult_Success) {
      break;
    }
    
    /* Merge in new data */
    memcpy(saveBuff + (start - blockStart), data, chunk);
    
    /* Erase the block */
    result = eraseFlash(pDevCtx, blockStart, blockSize);
    if (result != MethodResult_Success) {
      break;
    }
    
    /* Write the block */
    result = writeBlock(pDevCtx, blockStart, blockSize, saveBuff);
    if (result != MethodResult_Success) {
      break;
    }
    
    length -= chunk;
    start += chunk;
    data = (const void*) (((os_uint8*) data) + chunk);
  }
  
  return result;
}
