/*
** ADM-XRC Series Driver
**
** (C) Copyright 2001-2003 Alpha Data
**
** intwait.c - IOCTL handlers for waiting on FPGA interrupts.
**
** ----------------------------------------------------------------------------
*/

#include "platform.h"
#include "device.h"
#include "ioctl.h"
#include "intwait.h"

static void
intWaitCancel(
  OS_DEVICE_OBJECT* devObj,
  OS_IO_REQUEST*    req)
{
  OS_SPINLOCK_FLAGS f;
  DeviceContext*    devCtx = (DeviceContext*) os_get_device_context(devObj);
  os_boolean        complete;

  os_dbgprint(3, ("intWaitCancel: entered, req = 0x%p\n", req));

  os_acquire_spinlock(&devCtx->intwait.lock, &f);
  if (devCtx->intwait.waiting.req != NULL) {
    if (devCtx->intwait.waiting.completing) {
      /* Was about to be completed by intWaitComplete() */
      complete = TRUE;
    } else {
      /* Queued, intWaitComplete() not running */
      devCtx->intwait.waiting.req = NULL;
      complete = TRUE;
    }
  } else {
    os_dbgprint(3, ("intWaitCancel: no request queued, req = 0x%p\n", req));  
    complete = FALSE;
  }
  os_release_spinlock(&devCtx->intwait.lock, f);

  if (complete) {
    os_dbgprint(3, ("intWaitCancel: completing, req = 0x%p\n", req));  
    req->status.status = OS_MAKE_STATUS(ADMXRC_CANCELLED);
    os_complete_request(req);
  }
}

static void
intWaitComplete(
  DeviceContext* devCtx)
{
  OS_IO_REQUEST*                  req = NULL;
  unsigned long                   now;
  long                            remaining;
  OS_SPINLOCK_FLAGS               f;
  ADMXRC_IOCTLS_WAITFORINTERRUPT* optr;
  os_boolean                      complete;

  os_dbgprint(3, ("intWaitComplete: entered\n"));

  os_acquire_spinlock(&devCtx->intwait.lock, &f);
  req = devCtx->intwait.waiting.req;
  if (req == NULL) {
    /* No IO request waiting - set the signalled flag and do nothing more */
    os_dbgprint(6, ("intWaitComplete: no IO request queued, setting signalled flag\n"));
    devCtx->intwait.signalled = TRUE;
    complete = FALSE;
  } else {
    /* There is a waiting IO request */
    devCtx->intwait.waiting.req = NULL;
    devCtx->intwait.waiting.completing = TRUE;
    if (os_set_cancel_routine(req, NULL)) {
      /*
      ** Cancel routine wasn't running - so it's safe to complete
      ** the request.
      */
      complete = TRUE;
    } else {
      os_dbgprint(3, ("intWaitComplete: cancel routine running - not completing req 0x%p\n", req));
      /*
      ** Cancel routine is already running, but spinning on the lock.
      ** Let the cancel rountine complete the request.
      */
      complete = FALSE;
    }
  }
  os_release_spinlock(&devCtx->intwait.lock, f);

  if (complete) {
    os_dbgprint(3, ("intWaitComplete: completing req 0x%p\n", req));
    /*
    ** Cancel routine successfully removed - it's safe to complete
    ** the request here
    */
    now = os_get_time();
    optr = (ADMXRC_IOCTLS_WAITFORINTERRUPT*) req->param.outBuf;
    if (!devCtx->intwait.expire.never) {
      if (devCtx->intwait.expire.ticks > now) {
	remaining = devCtx->intwait.expire.ticks - now;
      } else {
	remaining = 0;
      }
#if 0
      optr->out.remaining_us = (unsigned long) (((os_uint64) remaining) * 1000000 / os_ticks_per_second);
#else
      optr->out.remaining_us = 0;
#endif
    }
    req->status.length = sizeof(optr->out);
    req->status.status = OS_STATUS_SUCCESS;
    os_complete_request(req);
  }
}

#if defined(ADMXRC_DRIVER_WINDOWS) || defined(ADMXRC_DRIVER_VXWORKS)
static void
signalIntEvents(
  DeviceContext* devCtx)
{
  /* Signal all registered interrupt wait events */
  KeAcquireSpinLock(&pDE->intEventLock, &irql);
  ev = pDE->intEventList;
  while (1) {
    if (ev == NULL) {
      break;
    }
    os_dbgprint((6, "signalIntEvents: event list node = 0x%lx\n",
		 (unsigned long) ev));
    os_set_event(ev->event, 0, FALSE);
    ev = ev->next;
  }
  KeReleaseSpinLock(&pDE->intEventLock, irql);
}
#endif

/*
** The FPGA interrupt DPC
*/

void
ADMXRC_FpgaIntDpc(
  void* context,
  void* argument)
{
  DeviceContext* devCtx = (DeviceContext*) context;

  intWaitComplete(devCtx);
#if defined(ADMXRC_DRIVER_WINDOWS) || defined(ADMXRC_DRIVER_VXWORKS)
  signalIntEvents(devCtx);
#endif
}

/*
** IOCTL handlers
*/

OS_STATUS
ADMXRC_CancelWaitForInterrupt(
  DeviceContext* devCtx,
  OS_IO_REQUEST* req)
{
  OS_SPINLOCK_FLAGS               f;
  os_boolean                      complete;
  OS_IO_REQUEST*                  creq;
  OS_STATUS                       status;

  os_dbgprint(3, ("ADMXRC_CancelWaitForInterrupte: entered\n"));

  os_acquire_spinlock(&devCtx->intwait.lock, &f);
  creq = devCtx->intwait.waiting.req;
  if (creq == NULL) {
    /* No IO request waiting - do nothing more */
    complete = FALSE;
  } else {
    /* There is a waiting IO request */
    devCtx->intwait.waiting.req = NULL;
    devCtx->intwait.waiting.completing = TRUE;
    if (os_set_cancel_routine(creq, NULL)) {
      /*
      ** Cancel routine wasn't running - so it's safe to complete
      ** the request.
      */
      complete = TRUE;
    } else {
      os_dbgprint(3, ("intWaitComplete: cancel routine running - not completing req 0x%p\n", req));
      /*
      ** Cancel routine is already running, but spinning on the lock.
      ** Let the cancel rountine complete the request.
      */
      complete = FALSE;
    }
  }
  os_release_spinlock(&devCtx->intwait.lock, f);

  if (complete) {
    /*
    ** Cancel routine successfully removed - it's safe to complete
    ** the request here
    */
    os_dbgprint(3, ("ADMXRC_CancelWaitForInterrupt: completing req 0x%p\n", req));
    creq->status.status = OS_MAKE_STATUS(ADMXRC2_CANCELLED);
    os_complete_request(creq);
    status = OS_STATUS_SUCCESS;
  } else {
    status = OS_MAKE_STATUS(ADMXRC2_FAILED);
  }

  return status;
}

OS_STATUS
ADMXRC_WaitForInterrupt(
  DeviceContext* devCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_WAITFORINTERRUPT* iptr;
  ADMXRC_IOCTLS_WAITFORINTERRUPT* optr;
  OS_SPINLOCK_FLAGS               f;
  unsigned long                   offset;
  OS_STATUS                       status;

  iptr = (ADMXRC_IOCTLS_WAITFORINTERRUPT*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(iptr->in)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }
  optr = (ADMXRC_IOCTLS_WAITFORINTERRUPT*) req->param.outBuf;
  if (optr == NULL || req->param.outSize != sizeof(optr->out)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  optr->out.expired = 0;
  os_acquire_spinlock(&devCtx->intwait.lock, &f);
  if (devCtx->intwait.waiting.req != NULL) {
    /* A thread is already waiting for an interrupt */
    os_release_spinlock(&devCtx->intwait.lock, f);
    status = OS_MAKE_STATUS(ADMXRC_DEVICE_BUSY);
  } else {
    if (devCtx->intwait.signalled) {
      /* An FPGA interrupt has already occurred */
      devCtx->intwait.signalled = FALSE;
      os_release_spinlock(&devCtx->intwait.lock, f);
      optr->out.remaining_us = iptr->in.timeout_us;
      req->status.length = sizeof(optr->out);
      status = OS_STATUS_SUCCESS;
    } else {
      if (os_set_cancel_routine(req, intWaitCancel)) {
	/* We must wait for an FPGA interrupt */
	if (0 && iptr->in.flags & ADMXRC_INTWAIT_TIMEOUT) {
	  offset = (unsigned long) (((os_uint64) iptr->in.timeout_us) * os_ticks_per_second / 1000000);
	  devCtx->intwait.expire.ticks = os_get_time() + offset;
	  devCtx->intwait.expire.never = FALSE;
	} else {
	  devCtx->intwait.expire.never = TRUE;
	}
	devCtx->intwait.waiting.req = req;
	devCtx->intwait.waiting.completing = FALSE;
	os_release_spinlock(&devCtx->intwait.lock, f);
	status = OS_STATUS_PENDING;
      } else {
	/* IO request already pending cancellation */
	os_release_spinlock(&devCtx->intwait.lock, f);
	status = OS_MAKE_STATUS(ADMXRC_CANCELLED);
      }
    }
  }

  return status;
}
